module Board
  ( Board
  , Cell(..)
  , Pos(..)
  , Dim(..)
  , dimensions
  , newBoard
  , addLine
  , (?)
  , shortestPath
  , dist
  )
 where

import Commands
import Foreign.Ptr
import Foreign.ForeignPtr
import Foreign.C.String
import System.IO.Unsafe( unsafePerformIO )

-- position, dimensions

data Pos = Pos !Int !Int deriving (Show, Eq, Ord)
data Dim = Dim !Int !Int deriving (Show, Eq, Ord)

dist :: Pos -> Pos -> Int
dist (Pos x1 y1) (Pos x2 y2) = abs (x1-x2) + abs (y1-y2)

-- cell

data Cell
  = Wall
  | Water
  | Base
  | Empty
 deriving ( Eq )

instance Show Cell where
  show Wall  = "#"
  show Water = "~"
  show Base  = "@"
  show Empty = "."

-- board implementation

data CBoard = CBoard
data Board  = Board Dim (ForeignPtr CBoard)

-- primitive C functions

foreign import ccall boardCreate       :: Int -> Int -> IO (Ptr CBoard)
foreign import ccall boardFree         :: Ptr CBoard -> IO ()
foreign import ccall boardAddLine      :: Ptr CBoard -> Int -> CString -> IO Int
foreign import ccall boardRead         :: Ptr CBoard -> Int -> Int -> IO Char
foreign import ccall boardShortestPath :: Ptr CBoard -> Int -> Int -> Int -> Int -> IO Int
foreign import ccall boardGetFoundPath :: Ptr CBoard -> IO CString

-- Haskell wrappers

newBoard :: Dim -> IO Board
newBoard dim@(Dim x y) =
  do pb  <- boardCreate x y
     fpb <- newForeignPtr pb (boardFree pb)
     return (Board dim fpb)

addLine :: Board -> Int -> String -> IO ()
addLine (Board dim fpb) y s =
  withForeignPtr fpb (\pb ->
    withCString s (\cs ->
      do boardAddLine pb y cs
         return ()
    )
  )

(?) :: Board -> Pos -> Cell
Board dim fpb ? Pos x y =
  unsafePerformIO (
    withForeignPtr fpb (\pb ->
      do c <- boardRead pb x y
         return (case c of
                   '#' -> Wall
                   '~' -> Water
                   '@' -> Base
                   '.' -> Empty
                )
    )
  )

shortestPath :: Board -> Pos -> Pos -> Maybe (Int,[Direction])
shortestPath (Board (Dim x y) fpb) (Pos xs ys) (Pos xt yt) =
  unsafePerformIO (
    withForeignPtr fpb (\pb ->
      do n <- boardShortestPath pb xs ys xt yt
         if n < 0
           then do return Nothing
           else do cs <- boardGetFoundPath pb
                   s  <- peekCString cs
                   return (Just (n, map dir (reverse s)))
    )
  )
 where
  dir 'N' = North
  dir 'E' = East
  dir 'S' = South
  dir 'W' = West

dimensions :: Board -> Dim
dimensions (Board dim _) = dim

-- example use:
main =
  do b <- newBoard (Dim 4 4)
     addLine b 4 "...."
     addLine b 3 "###."
     addLine b 2 ".#.."
     addLine b 1 "...."
     print (shortestPath b (Pos 1 1) (Pos 2 4))
