module Communicate where

import Network
import IO
import Char
import Commands
import System
import Board
import Random

-----------------
-- Initialisation


init s p = do server <- connectTo s (PortNumber (fromInteger p))
	      --putStrLn "Connected to server" --debug
	      hSetBuffering server NoBuffering
	      hPutStrLn server "Player"
	      --putStrLn "Sent \"Player\" to server" --debug
	      dim <- getBoardSize server
	      --putStrLn (show dim) --debug
	      b <- newBoard dim
	      board <- getBoard server b dim
	      --putStrLn ("Board read : " ++ show board) -- debug
	      player@(id,_,_) <- getPlayer server
	      --putStrLn (show player) --debug
	      resp <- getResponse server id
	      --putStrLn (show resp) --debug
	      rand <- newStdGen
	      return (server,(dim,board,b,player,resp, rand))

getBoardSize h = do lst <- aux h
		    case lst of
		      [x,y] | all isDigit x && all isDigit y ->
			      return (Dim (mkInt x) (mkInt y))
		      _ -> error "error in getBoardSize"
  where aux h = do line <- hGetLine h
		   return (words line)

mkInt str = foldl (\x y -> 10*x + fromEnum y - fromEnum '0' ) 0 str

getBoard h p (Dim x y) = mapM (getRow h p) [1 .. y]
  where getRow h p n = do line <- hGetLine h
			  addLine p n line
			  --putStrLn line
			  return line

getPlayer h = do line <- hGetLine h
		 case words line of
		      [x,y,z] | all isDigit x &&
				all isDigit y &&
				all isDigit z ->
				  return (mkInt x,mkInt y, mkInt z)
		      _ -> error ("Error in getPlayer : " ++ line)

------------------
-- Server response

getResponse h id = do b <- hIsEOF h
		      if b then exitWith ExitSuccess
		            else do line <- hGetLine h
				    --putStrLn line
				    return (getRobots id line)

getRobots id line = let list = map mkResponse (splitByRobots [] (words line))
		    in (expectJust (lookup id list),filter ((/=id).fst) list)

expectJust (Just a) = a
expectJust _ = error "Error in expectJust"

mkResponse robot = let (id,rest) = readRobotId robot
		   in (id,readActions rest)

readRobotId (('#':id):str) = (mkInt id,str)
readRobotId _ = error "Error in readRobotId"

readActions ("N":str) = Move North : readActions str
readActions ("S":str) = Move South : readActions str
readActions ("E":str) = Move East : readActions str
readActions ("W":str) = Move West : readActions str
readActions ("X":x:"Y":y:str) = Appear (mkInt x) (mkInt y) : readActions str
readActions ("P":id:str) = Pick [mkInt id] : readActions str
readActions ("D":id:str) = Drop [mkInt id] : readActions str
readActions _ = []

splitByRobots [] (r@('#':id):str) = splitByRobots [r] str
splitByRobots list (r@('#':id):str) = reverse list : splitByRobots [r] str
splitByRobots list (foo:str) = splitByRobots (foo:list) str
splitByRobots [] [] = []
splitByRobots list [] = [reverse list]

------------------
-- Sending Actions

sendAction h action = hPutStrLn h (show action)

---------------------------------
-- Receiving package descriptions

getPackages h = do b <- hIsEOF h
		   if b then exitWith ExitSuccess
		        else do line <- hGetLine h
				return (mkPackages line)

mkPackages list = gather (map mkInt (words list))

gather (id:x:y:w:rest) = (id,x,y,w) : gather rest
gather _ = []
