module GameState where

import Commands
import Board
import Foreign.Ptr
import Data.FiniteMap
import Data.List ((\\))
import Maybe
import Random

----------------
-- A State Monad

data State = MkState { -- Static
		       board               :: Board ,
		       boardSize           :: Dim ,
		       myId                :: !Int ,
		       wholeCapacity       :: !Int ,

                       --Dynamic
                       score               :: !Int ,
                       money               :: !Int ,
		       pickablePacks       :: [Package] ,
		       myPos               :: !Pos ,
		       myCapacity          :: !Int ,
		       myPacks             :: [Package] ,
		       certainPacks        :: FiniteMap PackageID (Pos,Package) ,
		       maybePacks          :: FiniteMap PackageID Pos ,
		       otherBots           :: FiniteMap RobotID Pos ,
		       opponents           :: [BotInfo] ,
		       notVisitedHomeBases :: [Pos] ,
                       randomSeed          :: StdGen
		     }

instance Show State where
  show state = show (notVisitedHomeBases state)

showState state = --"\ESC[2J"
		  at (1,1) boardAsString
		  ++ bots ++ me ++ goto 1 (y+1) ++ "\ESC[0J"
  where boardAsString = unlines (map
				 (\b -> concatMap
				  (\a -> show (theBoard ? (Pos a b)))
				   [1..x] )
				  [y,y-1..1])

	theBoard = board state
	Dim x y = boardSize state
	bots = concatMap (\ (Pos a b) -> at (a,y - b + 1) "X")
	                 (eltsFM $ otherBots state)
	Pos xm ym = myPos state
	me = at (xm, y - ym + 1) "O"


	at (x,y) s  = goto x y ++ s
	highlight s = "\ESC[7m"++s++"\ESC[0m"
	goto x y    = '\ESC' : '[':(show y ++(';':show x ++ "H"))
	home        = goto 1 1

-- id x y weight
data Package = Pack { pid     :: !Int,
		      packPos :: !Pos,
		      pw      :: !Int
		      } deriving (Eq,Show)

data BotInfo = BotInfo { botId    :: !RobotID ,
			 botPos   :: !Pos ,
			 botPacks :: [PackageID]
			 }

newtype SM a = SM (State -> (a , State))
unSM (SM f) = f

instance Monad SM where
  return a = SM (\s -> (a,s))
  (SM f) >>= g = SM (\s -> case f s of
		             (a,s') -> unSM (g a) s')

update f = SM (\s -> ((),f s))

peekWith f = SM (\s -> (f s,s))

runSM (SM f) state = f state

-- Init state

mkInitState (d@(Dim x y),board,p,(id,capacity,money),resp, rand)
    = MkState { boardSize = d,
		wholeCapacity = capacity,
	        money = money,
		score = 0,
		pickablePacks = [],
                myId = id,
                myPos = (getMyPos resp),
	        myCapacity = capacity,
	        myPacks = [],
	        certainPacks = emptyFM,
	        maybePacks = emptyFM,
	        otherBots = (getOtherBots resp),
		opponents = getOpponents resp,
	        notVisitedHomeBases = homeBases board,
	        board = p,
                randomSeed = rand
	        }


getMyPos ([Appear x y],_) = Pos x y
getMyPos _ = error "Error in getMyPos"

getOtherBots (_,resp) = listToFM (map mkBot resp)
  where mkBot (id,[Appear x y]) = (id,Pos x y)
	mkBot _ = error "Error in mkBot"

getOpponents (_,resp) = map mkBotInfo resp
  where mkBotInfo (id,[Appear x y]) = BotInfo { botId = id ,
						botPos = Pos x y ,
						botPacks = []
						}
	mkBotInfo _ = error "Error in mkBotInfo"

homeBases board = [ Pos x y | (y,row) <- zip [1..] board
		            , (x,c) <- zip [1..] row
		            , c == '@']

--------------
-- Some action

mkAction state = C 1 (Move North)

------------------------------------------------------------------
-- When we get to know the packages in the beginning of the round, 
-- store that information in the state

considerPacks state pkg
    = state{ maybePacks = filterFM isStillMaybe (maybePacks state) ,
--	     delListFromFM (maybePacks state) (map getId pkg),
	     certainPacks = addListToFM (certainPacks state) (map mkPackId pkg),
	     pickablePacks = map mkPack pkg,
             randomSeed = snd (next (randomSeed state))
	     }
  where getId (id,_,_,_) = id
	pos = myPos state
	mkPackId (id,x,y,w) = (id, (pos,Pack id (Pos x y) w))
	mkPack (id,x,y,w) = (Pack id (Pos x y) w)
	packids = map getId pkg
	isStillMaybe id ppos = not (id `elem` packids)
			       && ppos /= pos

----------------------------------------------------------
-- Here is where we update the state at the end of a turn.

newState (myActions,bots) state
    = checkHomeBases $
      removeBots bots $
      foldl botupd
	    (foldl myupd state myActions) bots

checkHomeBases state
    = state{ notVisitedHomeBases = filter (/= myPos state)
	                           (notVisitedHomeBases state)
	     }

removeBots bots state
    = state{ otherBots = intersectFM_C const (otherBots state)
	                                     (listToFM bots),
	     opponents = filter (\ bot -> maybeToBool
				 (lookupFM botsFM (botId bot))
				 ) (opponents state)
	     }
  where botsFM = listToFM bots
        maybeToBool = maybe False (const True)

-- Updating my own data

myupd state (Move d) = state{ myPos = move d (myPos state) }
myupd state (Pick pids)
    = let oldPacks = myPacks state
	  newPacks = concatMap (\pid -> maybeToList $
			        lookupFM (certainPacks state) pid) pids
      in state{ myPacks = oldPacks ++ map snd newPacks,
		certainPacks = delListFromFM (certainPacks state) pids,
		myCapacity = (myCapacity state) - sum (map (pw.snd) newPacks)
		}
myupd state (Drop pids)
    = let oldPacks = myPacks state
	  pos = myPos state
	  dropped = filter (hasIds pids) oldPacks
	  delivered = filter ((== pos).packPos) dropped
	  notDelivered = filter ((/= pos).packPos) dropped
      in state{ myPacks = filter (not . hasIds pids) oldPacks,
		certainPacks = addListToFM (certainPacks state)
		                (map tagId (zip (repeat pos) notDelivered)),
		myCapacity = (myCapacity state) + sum (map pw dropped),
		score = (score state) + sum (map pw delivered)
		}
myupd state (Appear x y) = state{ myPos = Pos x y }


-- updating which depends on what the other bots are doing

botupd state (id,actions)
    = foldl (upd id) state actions

upd id state (Move d)
    = state{ otherBots = addToFM (otherBots state) id (move d oldPos),
	     opponents = changeBot id (moveBot d) (opponents state)
	    }
  where oldPos = case lookupFM (otherBots state) id of
	              Just p -> p
		      _ -> error "Error in upd case Move"
upd id state (Pick pids)
    = state{ certainPacks = delListFromFM (certainPacks state) pids,
	     opponents = changeBot id (addPacks pids) (opponents state)
	     }
upd id state (Drop pids)
    = state{ maybePacks = addListToFM (maybePacks state)
	                              (zip pids (repeat pos)),
	     opponents = changeBot id (removePacks pids) (opponents state)
	     }
  where pos = case lookupFM (otherBots state) id of
	           Just p -> p
		   _ -> error "Error in upd case Drop"
upd id state (Appear x y)
    = state{ otherBots = addToFM (otherBots state) id (Pos x y),
	     opponents = changeBot id (setPos x y) (opponents state)
	    }

-- aux

changeBot id f bots = map (\bot -> if botId bot == id then f bot else bot) bots
addPacks pids bot = bot{ botPacks = (botPacks bot) ++ pids }
removePacks pids bot = bot{ botPacks = (botPacks bot) \\ pids }
setPos x y bot = bot{ botPos = Pos x y }

hasIds ids pack = pid pack `elem` ids

tagId t@(_,pack) = (pid pack,t)

move North (Pos x y) = Pos x (y+1)
move South (Pos x y) = Pos x (y-1)
move East  (Pos x y) = Pos (x+1) y
move West  (Pos x y) = Pos (x-1) y

moveBot d bot = bot{ botPos = move d (botPos bot) }
