module Memo
  -- memo functions
  ( Memo    -- :: class
    ( memo  -- :: (a -> b) -> (a -> b)
    )
  )
 where

{-

This module provides a type class interface to "memo functions", and
provides instances for the standard Haskell types: (), (,), Either,
Maybe, [], Int, Integer, Bool.

A "memo function" is a function that remembers the results of of all its
previous calls, and reuses them when called again with the same argument
as before.

It is simple to use: When given a function f, and defining "g = memo f", one
gets a function g that behaves exactly the same as f, but it is guaranteed
that evaluating "g x" is *identical* to evaluating "g x".

Its implementation works by keeping track of a little lookup table (often
called "memotable"), which is has constant lookup-time for finite types (such
as Bool), and logarithmic lookup-time for infinite types (such as Int). The
memotable is completely implicit in the declarations below however. Also, the
memotable is a static thing, no dynamic update is necessary because of
laziness.

Caution:

  * Memory problems: If you keep a reference to the function "g", the memotable
    stays in memory too!

  * When memoizing recursive functions, there is a big difference between
    memoizing at every recursive call, memoizing only calls from outside,
    or doing both.

-}

import Char

------------------------------------------------------------------
-- memoizable types

class Memo a where
  memo :: (a -> b) -> (a -> b)

------------------------------------------------------------------
-- instances

instance Memo () where
  memo f = \() -> fUnit
   where
    fUnit = f ()
    
instance Memo Bool where
  memo f = \x -> if x then fTrue else fFalse
   where
    fTrue  = f True
    fFalse = f False

instance (Memo a, Memo b) => Memo (a,b) where
  memo f = \(x,y) -> fPair x y
   where
    fPair = memo (\x -> memo (\y -> f (x,y)))

instance (Memo a, Memo b) => Memo (Either a b) where
  memo f = \x -> case x of
                   Left  a -> fLeft a
                   Right b -> fRight b
   where
    fLeft  = memo (f . Left)
    fRight = memo (f . Right)

instance Memo a => Memo (Maybe a) where
  memo f = \x -> case x of
                   Nothing -> fNothing
                   Just x  -> fJust x
   where
    fNothing = f Nothing
    fJust    = memo (f . Just)

instance Memo a => Memo [a] where
  memo f = \x -> case x of
                   []   -> fNil
                   a:as -> fCons a as
   where
    fNil  = f []
    fCons = memo (\a -> memo (\as -> f (a:as)))

instance Memo Integer where
  memo f = \x -> if x >= 0 then fPositive x else fNegative x
   where
    fPositive = memoNat f
    fNegative = memoNat (f . negate) . negate

instance Memo Int where
  memo f = \x -> if x >= 0 then fPositive x else fNegative x
   where
    fPositive = memoNat f
    fNegative = memoNat (f . negate) . negate

instance Memo Char where
  memo f = memo (f . chr) . ord
  
------------------------------------------------------------------
-- helpers

memoNat :: Integral a => (a -> b) -> (a -> b)
memoNat f = \x -> case x of
                    0             -> fZero
                    _ | even x    -> fEven (x `div` 2)
                      | otherwise -> fOdd  (x `div` 2)
 where
  fZero = f 0
  fEven = memoNat (\m -> f (2*m))
  fOdd  = memoNat (\m -> f (2*m + 1))

------------------------------------------------------------------
-- the end.

    
    
