/* --------------------------------------------------------------------------
   -- preamble
 */

#include <stdio.h>
#include <string.h>

#define INFINITY 3000001  /* > maximal path length */

/* --------------------------------------------------------------------------
   -- type definitions
 */

typedef enum
{ UNKNOWN, STAY, NORTH, EAST, SOUTH, WEST
} Direction;

typedef struct
{
  int xsize, ysize;
  int xsource, ysource;
  int xtarget, ytarget;
  int foundLen;
  int foundWeight;
  char      **map;
  int       **weight;
  Direction **dir;
  char *path;
} Board;

/* --------------------------------------------------------------------------
   -- function headers
 */

Board *boardCreate( int x, int y );
void   boardFree( Board *b );
int    boardAddLine( Board *b, int x, char *line );

char   boardRead( Board *b, int x, int y );

int    boardShortestPath( Board *b, int xs, int ys, int xt, int yt );
char  *boardGetFoundPath( Board *p );

/* --------------------------------------------------------------------------
   -- implementation
 */

Board *boardCreate( int x, int y )
{
  Board *b = (Board*)malloc(sizeof(Board));
  int i;

  b->xsize   = x;
  b->ysize   = y;
  b->xsource = -1;
  b->ysource = -1;

  b->map    = (char**)malloc((y+2) * sizeof(char*));
  b->weight = (int**)malloc((y+1) * sizeof(int*));
  b->dir    = (Direction**)malloc((y+1) * sizeof(Direction*));
  
  for ( i = 1; i <= y; i++ )
  {
    b->map[i]    = (char*)malloc((x+2) * sizeof(char));
    b->weight[i] = (int*)malloc((x+1) * sizeof(int));
    b->dir[i]    = (Direction*)malloc((x+1) * sizeof(Direction));
  } 

  b->path = NULL;
  return b;
}

void boardFree( Board *b )
{
  int i;
  
  for ( i = 1; i <= b->ysize; i++ )
  {
    free( b->map[i] );
    free( b->weight[i] );
    free( b->dir[i] );
  }
  
  free( b->map );
  free( b->weight );
  free( b->dir );
  
  if ( b->path )
    free( b->path );
  
  free( b );
}

int boardAddLine( Board *b, int y, char *line )
{
  int j;

  for ( j = 1; j <= b->xsize; j++ )
    b->map[y][j] = line[j-1];

  return 1;
}

char boardRead( Board *b, int x, int y )
{
  return b->map[y][x];
}

int abs( int x )
{
  return x < 0 ? -x : x;
}

void boardFind( Board *b, int x, int y, Direction dir, int weight, int len )
{
  if ( x < 1
    || y < 1
    || x > b->xsize
    || y > b->ysize
    || b->map[y][x] == '~'
    || b->map[y][x] == '#'
    || weight + abs(x - b->xtarget) + abs(y - b->ytarget) >= b->foundWeight
    || b->weight[y][x] <= weight
     )
    return;

  if ( (y > 1 && b->map[y-1][x] == '~')
    || (x > 1 && b->map[y][x-1] == '~')
    || (y < b->ysize && b->map[y+1][x] == '~')
    || (x < b->xsize && b->map[y][x+1] == '~')
     )
    weight+=5;

  b->dir[y][x]    = dir;
  b->weight[y][x] = weight;

  if ( x == b->xtarget && y == b->ytarget )
  {
    b->foundLen    = len;
    b->foundWeight = weight;
  }
  else
  {
    int i;
    Direction next[4];
    
    if ( x < b->xtarget && y <= b->ytarget )
    { next[0] = EAST;  next[1] = NORTH; next[2] = WEST;  next[3] = SOUTH; }
    else if ( x >= b->xtarget && y < b->ytarget )
    { next[0] = NORTH; next[1] = WEST;  next[2] = SOUTH; next[3] = EAST;  }
    else if ( x > b->xtarget && y >= b->ytarget )
    { next[0] = WEST;  next[1] = SOUTH; next[2] = EAST;  next[3] = NORTH; }
    else if ( x <= b->xtarget && y > b->ytarget )
    { next[0] = SOUTH; next[1] = WEST;  next[2] = NORTH; next[3] = EAST;  }
    
    weight++;
    len++;

    for ( i = 0; i < 4; i++ )
      switch ( next[i] )
      {
        case NORTH: boardFind( b, x, y+1, next[i], weight, len ); break;
        case EAST:  boardFind( b, x+1, y, next[i], weight, len ); break;
        case SOUTH: boardFind( b, x, y-1, next[i], weight, len ); break;
        case WEST:  boardFind( b, x-1, y, next[i], weight, len ); break;
      }
  }
}

int boardShortestPath( Board *b, int xs, int ys, int xt, int yt )
{
  int i, j, x, y;

  if ( 1 /* xs != b->xsource || ys != b->ysource */ )
  {
    for ( i = 1; i <= b->ysize; i++ )
      for ( j = 1; j <= b->xsize; j++ )
      {
        b->weight[i][j] = INFINITY;
        b->dir[i][j]    = UNKNOWN;
      }

    b->xsource = xs;
    b->ysource = ys;
  }

  b->xtarget     = xt;
  b->ytarget     = yt;
  b->foundLen    = INFINITY;
  b->foundWeight = INFINITY;
  
  boardFind( b, xs, ys, STAY, 0, 0 );

  if ( b->foundLen == INFINITY )
    return (-1);
  
  b->path = (char*)realloc( b->path, b->foundLen+1 );
  
  x = xt;
  y = yt;
  for ( i = 0; x != xs || y != ys; i++ )
  {
    switch ( b->dir[y][x] )
    {
      case NORTH: b->path[i] = 'N'; y--; break;
      case EAST:  b->path[i] = 'E'; x--; break;
      case SOUTH: b->path[i] = 'S'; y++; break;
      case WEST:  b->path[i] = 'W'; x++; break;
    }
  }

  b->path[i] = '\0';
  return b->foundWeight;
}

char *boardGetFoundPath( Board *b )
{
  return b->path;
}

/* --------------------------------------------------------------------------
   -- main
 */

main_()
{
  Board *b = boardCreate( 21, 21 );
  int n;
  
  boardAddLine( b, 1, "....................." );
  boardAddLine( b, 2, ".........#.#........." );
  boardAddLine( b, 3, ".........#.#........." );
  boardAddLine( b, 4, ".........#.#........." );
  boardAddLine( b, 5, ".........#.#........." );
  boardAddLine( b, 6, ".........#.#........." );
  boardAddLine( b, 7, ".........#.#........." );
  boardAddLine( b, 8, "......####.####......" );
  boardAddLine( b, 9, "~~~~~~#.......#~~~~~~" );
  boardAddLine( b, 10, "#~#~#~#..@@@..#~#~#~#" );
  boardAddLine( b, 11, ".........@@@........." );
  boardAddLine( b, 12, ".........@@@........." );
  boardAddLine( b, 13, "~#~#~##..@@@..##~#~#~" );
  boardAddLine( b, 14, "~~~~~~#.......#~~~~~~" );
  boardAddLine( b, 15, "~~~~~~####.####~~~~~~" );
  boardAddLine( b, 16, "~~~~~~~~~#.#~~~~~~~~~" );
  boardAddLine( b, 17, "~~~~~~~~~#.#~~~~~~~~~" );
  boardAddLine( b, 18, "~~~~~~~~~#.#~~~~~~~~~" );
  boardAddLine( b, 19, "~~~~~~~~~#.#~~~~~~~~~" );
  boardAddLine( b, 20, "~~~~~~~~~#.#~~~~~~~~~" );
  boardAddLine( b, 21, "~~~~~~~~~#.#~~~~~~~~~" );
  
  n = boardShortestPath( b, 1, 1, 10, 10 );
  if ( n >= 0 )
    printf( "Path (%d): [%s] (reversed!)\n", n, boardGetFoundPath( b ) );
  else
    printf( "No path found\n" );
  
  boardFree( b );
}

/* --------------------------------------------------------------------------
   -- the end.
 */
